import matplotlib.pyplot as plt
import numpy as np
from scipy.sparse.csgraph import reverse_cuthill_mckee


class SparseMatrixGrapher:
    def __init__(self, matrix):
        self.matrix = matrix

    def graph(
        self,
        filename=None,
        marker_string="c.",
        mew=0.5,
        ms=1,
        alpha=0.8,
        width=None,
        height=None,
        dpi=300,
    ):
        tm = self.matrix.tocoo()
        y, x = self.matrix.shape
        plt.figure(figsize=(width or x / 1000, height or y / 1000))
        ax = plt.axes([0, 0, 1, 1])
        # Start from top left corner, not bottom left corner
        ax.xaxis.set_ticks_position("none")
        ax.yaxis.set_ticks_position("none")
        ax.xaxis.set_ticklabels([])
        ax.yaxis.set_ticklabels([])
        ax.plot(
            tm.shape[1] - tm.col, tm.row, marker_string, mew=mew, ms=ms, alpha=alpha
        )
        ax.set_ylim((0, tm.shape[0]))
        ax.set_xlim((0, tm.shape[1]))
        plt.box(False)
        if filename:
            plt.savefig(filename, dpi=dpi)

    def magnitude_graph(self, filename=None, dpi=600, width=None, height=None):
        def get_relative_scores(data):
            return np.abs(data) / np.max(np.abs(data))

        def get_colors(distances):
            cmap = plt.get_cmap("Dark2")
            return cmap(distances)

        def unroll(data):
            return [list(row) for row in data]

        nm = reverse_cuthill_mckee(self.matrix)
        ro = self.matrix[nm, :][:, nm]
        as_coo = ro.tocoo()
        y, x = as_coo.shape

        colors = unroll(get_colors(get_relative_scores(as_coo.data)))

        plt.figure(figsize=(width or x / 1000, height or y / 1000))
        ax = plt.axes([0, 0, 1, 1])
        plt.scatter(
            list(as_coo.shape[1] - as_coo.col),
            list(as_coo.row),
            s=10,
            c=colors,
            marker=".",
            edgecolors="None",
        )
        ax.xaxis.set_ticks_position("none")
        ax.yaxis.set_ticks_position("none")
        ax.xaxis.set_ticklabels([])
        ax.yaxis.set_ticklabels([])
        ax.set_ylim((0, self.matrix.shape[0]))
        ax.set_xlim((0, self.matrix.shape[1]))
        plt.box(False)
        if filename:
            plt.savefig(filename, dpi=dpi)

    def ordered_graph(self, filename=None, dpi=600, width=None, height=None):
        def get_distances(xs, ys):
            z = np.abs(xs - ys) / 2
            return np.sqrt(2 * z ** 2) / MAX_DIST

        def get_colors(distances):
            cmap = plt.get_cmap("Dark2")
            return cmap(distances)

        def unroll(data):
            return [list(row) for row in data]

        nm = reverse_cuthill_mckee(self.matrix)
        ro = self.matrix[nm, :][:, nm]
        as_coo = ro.tocoo()
        y, x = as_coo.shape
        MAX_DIST = np.sqrt(2 * (x / 2.0) ** 2)

        colors = unroll(get_colors(get_distances(as_coo.col, as_coo.row)))

        plt.figure(figsize=(width or x / 1000, height or y / 1000))
        ax = plt.axes([0, 0, 1, 1])
        plt.scatter(
            list(as_coo.shape[1] - as_coo.col),
            list(as_coo.row),
            s=10,
            c=colors,
            marker=".",
            edgecolors="None",
        )
        ax.xaxis.set_ticks_position("none")
        ax.yaxis.set_ticks_position("none")
        ax.xaxis.set_ticklabels([])
        ax.yaxis.set_ticklabels([])
        ax.set_ylim((0, self.matrix.shape[0]))
        ax.set_xlim((0, self.matrix.shape[1]))
        plt.box(False)
        if filename:
            plt.savefig(filename, dpi=dpi)



import os

import numpy as np
from bw2calc import LCA
from bw2data import Database, projects
from stats_arrays import (
    LognormalUncertainty,
    NormalUncertainty,
    TriangularUncertainty,
    UniformUncertainty,
    uncertainty_choices,
)

from .matrix_grapher import SparseMatrixGrapher
from .page_rank import PageRank


class DatabaseHealthCheck:
    def __init__(self, database):
        self.db = Database(database)
        self.db.filters = {"type": "process"}

    def check(self, graphs_dir=None):
        tg, tfn, bg, bfn = self.make_graphs(graphs_dir)
        aggregated = self.aggregated_processes()
        return {
            "tg": tg,
            "tfn": tfn,
            "bg": bg,
            "bfn": bfn,
            "pr": self.page_rank(),
            "ue": self.unique_exchanges(),
            "uncertainty": self.uncertainty_check(),
            "sp": aggregated["system_processes"],
            "me": aggregated["many_exchanges"],
            "nsp": self.no_self_production(),
            "mo": self.multioutput_processes(),
            "ob": {},
        }

    def make_graphs(self, graphs_dir=None):
        lca = LCA({self.db.random(): 1})
        lca.lci()
        tech_filename = self.db.filename + ".technosphere.png"
        tech_filepath = os.path.join(graphs_dir or projects.output_dir, tech_filename)
        SparseMatrixGrapher(lca.technosphere_matrix).graph(tech_filepath, dpi=600)
        bio_filename = self.db.filename + ".biosphere.png"
        bio_filepath = os.path.join(graphs_dir or projects.output_dir, bio_filename)
        SparseMatrixGrapher(lca.biosphere_matrix).graph(bio_filepath, dpi=600)
        return tech_filepath, tech_filename, bio_filepath, bio_filename

    def page_rank(self):
        return PageRank(self.db).calculate()

    def unique_exchanges(self):
        data = self.db.load()
        exchanges = [
            (exc["input"], exc["amount"], exc["type"])
            for ds in data.values()
            for exc in ds.get("exchanges", [])
            if exc["type"] in {"biosphere", "technosphere"}
        ]
        bio_exchanges = [obj for obj in exchanges if obj[2] == "biosphere"]
        tech_exchanges = [obj for obj in exchanges if obj[2] == "technosphere"]
        return (
            len(tech_exchanges),
            len(set(tech_exchanges)),
            len(bio_exchanges),
            len(set(bio_exchanges)),
        )

    def uncertainty_check(self):
        # TODO: Also report no (None) uncertainty
        data = self.db.load()
        results = {obj.id: {"total": 0, "bad": 0} for obj in uncertainty_choices}
        for ds in data.values():
            for exc in ds.get("exchanges", []):
                ut = exc.get("uncertainty type")
                if ut is None:
                    continue
                results[ut]["total"] += 1
                if ut == LognormalUncertainty.id:
                    right_amount = np.allclose(
                        np.log(np.abs(exc["amount"])), exc["loc"], rtol=1e-3
                    )
                    if not exc.get("scale") or not right_amount:
                        results[ut]["bad"] += 1
                elif ut == NormalUncertainty.id:
                    if not exc.get("scale") or abs(exc["amount"]) != exc["loc"]:
                        results[ut]["bad"] += 1
                elif ut in {TriangularUncertainty.id, UniformUncertainty.id}:
                    if exc["minimum"] >= exc["maximum"]:
                        results[ut]["bad"] += 1
        return results

    def multioutput_processes(self):
        num_production_exchanges = [
            (
                key,
                len(
                    [
                        exc
                        for exc in ds.get("exchanges")
                        if exc["type"] == "production" and exc["input"] != key
                    ]
                ),
            )
            for key, ds in self.db.load().items()
        ]
        return [obj for obj in num_production_exchanges if obj[1]]

    def aggregated_processes(self, cutoff=500):
        num_exchanges = {
            key: {
                "technosphere": len(
                    [
                        exc
                        for exc in value.get("exchanges", [])
                        if exc["type"] == "technosphere"
                    ]
                ),
                "biosphere": len(
                    [
                        exc
                        for exc in value.get("exchanges", [])
                        if exc["type"] == "biosphere"
                    ]
                ),
            }
            for key, value in self.db.load().items()
            if value.get("type", "process") == "process"
        }
        system_processes = [
            (key, value["biosphere"])
            for key, value in num_exchanges.items()
            if value["technosphere"] == 0 and value["biosphere"] > cutoff
        ]
        many_exchanges = [
            (key, value["technosphere"])
            for key, value in num_exchanges.items()
            if value["technosphere"] > cutoff
        ]
        return {"system_processes": system_processes, "many_exchanges": many_exchanges}

    def no_self_production(self):
        def self_production(ds):
            return any(
                exc.input == exc.output
                for exc in ds.get("exchanges", [])
                if exc["type"] in ("production", "generic production")
            )

        return {
            ds.key
            for ds in self.db
            if ds.get("type", "process") == "process" and not self_production(ds)
        }

    # def ouroboros(self):
    #     """Find processes that consume their own reference products as inputs. Not necessarily an error, but should be examined carefully (see `Two potential points of confusion in LCA math <http://chris.mutel.org/too-confusing.html>`__).

    #     Returns:
    #         A set of database keys.

    #     """
    #     return {
    #         key
    #         for key, value in self.db.load().items()
    #         if any(
    #             exc
    #             for exc in value.get("exchanges", [])
    #             if exc["input"] == key and exc["type"] == "technosphere"
    #         )
    #     }



from bw2calc import LCA
from numpy import absolute, array, dot, ones, where


class ConvergenceError(Exception):
    pass


class PageRank:
    def __init__(self, database):
        self.database = database

    def calculate(self):
        self.lca = LCA({self.database.random(): 1})
        self.lca.lci()
        self.matrix = self.lca.technosphere_matrix.transpose()
        self.pr = [
            (x[0], self.lca.dicts.activity.reversed[x[1]])
            for x in self.page_rank(self.matrix)
        ]
        return self.pr

    def page_rank(self, technosphere, alpha=0.85, max_iter=100, tol=1e-6):
        """
        Return the PageRank of the nodes in the graph.

        Adapted from http://networkx.lanl.gov/svn/networkx/trunk/networkx/algorithms/link_analysis/pagerank_alg.py

        PageRank computes a ranking of the nodes in the graph G based on
        the structure of the incoming links. It was originally designed as
        an algorithm to rank web pages.

        The eigenvector calculation uses power iteration with a SciPy
        sparse matrix representation.

        Args:
            * *technosphere* (scipy sparse matrix): The technosphere matrix.
            * *alpha* (float, optional): Damping parameter for PageRank, default=0.85

        Returns:
            * Dictionary of nodes (activity codes) with value as PageRank


        References

        .. [1] A. Langville and C. Meyer,
           "A survey of eigenvector methods of web information retrieval."
           http://citeseer.ist.psu.edu/713792.html
        .. [2] Page, Lawrence; Brin, Sergey; Motwani, Rajeev and Winograd, Terry,
           The PageRank citation ranking: Bringing order to the Web. 1999
           http://dbpubs.stanford.edu:8090/pub/showDoc.Fulltext?lang=en&doc=1999-66&format=pdf
        """
        mat = technosphere.copy()
        (n, m) = mat.shape
        assert n == m  # should be square
        nodelist = range(n)

        # Drop diagonals, and only indicate adjacency
        mat.data[:] = 1
        for x in range(n):
            mat[x, x] = 0

        column_sum = array(mat.sum(axis=1)).flatten()
        index = where(column_sum != 0)[0]
        mat = mat.tolil()
        for i in index:
            # Workaround for lack of fancy indexing in CSR matrices
            mat[i, :] *= 1.0 / column_sum[i]

        mat = mat.tocsc()
        x = ones((n)) / n  # initial guess
        dangle = array(where(mat.sum(axis=1) == 0, 1.0 / n, 0)).flatten()
        i = 0

        while True:  # power iteration: make up to max_iter iterations
            xlast = x
            x = alpha * (x * mat + dot(dangle, xlast)) + (1 - alpha) * xlast.sum() / n
            # check convergence, l1 norm
            err = absolute(x - xlast).sum()
            if err < n * tol:
                break
            if i > max_iter:
                raise ConvergenceError(
                    "pagerank: power iteration "
                    "failed to converge in %d iterations." % (i + 1)
                )
            i += 1

        return sorted(zip(x, nodelist), reverse=True)



__all__ = [
    "compare_activities_by_grouped_leaves",
    "compare_activities_by_lcia_score",
    "ContributionAnalysis",
    "DatabaseHealthCheck",
    "find_differences_in_inputs",
    "GTManipulator",
    "PageRank",
    "print_recursive_calculation",
    "print_recursive_supply_chain",
    # "SerializedLCAReport",
    "traverse_tagged_databases",
]

from .comparisons import (
    compare_activities_by_grouped_leaves,
    compare_activities_by_lcia_score,
    find_differences_in_inputs,
)
from .contribution import ContributionAnalysis
from .health_check import DatabaseHealthCheck
from .page_rank import PageRank

# from .report import SerializedLCAReport
from .sc_graph import GTManipulator
from .tagged import traverse_tagged_databases
from .utils import print_recursive_calculation, print_recursive_supply_chain
from .version import version as __version__



import numpy as np


def gini_coefficient(x):
    """
    Return computed Gini coefficient.

    See https://en.wikipedia.org/wiki/Gini_coefficient

    Adapted from econpy library.
    copyright: 2005-2009 Alan G. Isaac
    license: MIT license
    contact: aisaac AT american.edu

    Args:
        *x* (list or array): Data

    Returns:
        Gini coefficient (float)

    """
    x = np.array(x)
    x.sort()
    y = np.cumsum(x)
    length = float(x.shape[0])
    B = y.sum() / (y[-1] * length)
    return float(1.0 + 1 / length - 2 * B)


def herfindahl_index(x, normalize=True):
    """
    Return computed Herfindahl index.

    See https://en.wikipedia.org/wiki/Herfindahl_index

    Normalized scores are bounded [0, 1]; non-normalized scores are [1/len(x), 1]. Normalization only counts non-zero values.

    Args:
        *x* (list or array): Data
        *normalize* (bool, default=True): Flag to normalize scores.

    Returns:
        Herfindahl index (float)

    """
    # Normalize so that total is 1
    x = np.array(x) / np.sum(x)
    index = (x ** 2).sum()
    if normalize:
        correction = 1 / (x != 0).sum()
        index = (index - correction) / (1 - correction)
    return float(index)


def concentration_ratio(x, number=4):
    """
    Return computed concentration ratio.

    See https://en.wikipedia.org/wiki/Concentration_ratio

    The concentration ratio measures the share of the market controlled by the top *number* firms. Returned ratio values vary from 0 to 1.

    Args:
        *x* (list or array): Data
        *number* (int, default=4): Number of values to consider. 4 and 8 are commonly used.

    Returns:
        Concentration ratio (float)

    """
    # Normalize so that total is 1
    x = np.array(x) / np.sum(x)
    x.sort()
    return float(x[-number:].sum())


def theil_index(x):
    """
    Return Theil entropy index.

    See https://en.wikipedia.org/wiki/Theil_Index

    The Theil index is a measure of economic inequality based on information theory. It is the difference between a dataset's maximum possible entropy and observed entropy.

    Args:
        *x* (list or array): Data

    Returns:
        Theil index (float)

    """
    x = np.array(x).copy()
    # Have to filter values because of log transform
    x = np.abs(x[x != 0])
    average, n = np.average(x), x.shape[0]
    return float(1 / n * ((x / average) * np.log(x / average)).sum())



import os
import uuid

import numpy as np
import requests
from bw2calc import LCA, GraphTraversal, ParallelMonteCarlo
from bw2data import JsonWrapper, config, get_activity, methods, projects
from scipy.stats import gaussian_kde

from .contribution import ContributionAnalysis
from .econ import concentration_ratio, herfindahl_index
from .sc_graph import GTManipulator


class SerializedLCAReport:
    """A complete LCA report (i.e. LCA score, Monte Carlo uncertainty analysis, contribution analysis) that can be serialized to a defined standard."""

    version = 2

    def __init__(self, activity, method, iterations=10000, cpus=None, outliers=0.025):
        self.activity = activity
        self.method = method
        self.iterations = iterations
        self.cpus = cpus
        self.outliers = outliers
        self.uuid = uuid.uuid4().hex

    def calculate(self):
        """Calculate LCA report data"""
        lca = LCA(self.activity, self.method)
        lca.lci()
        lca.lcia()

        gt = GraphTraversal().calculate(self.activity, method=self.method)
        print("FD")
        force_directed = self.get_force_directed(gt["nodes"], gt["edges"], lca)
        print("CA")
        ca = ContributionAnalysis()
        print("hinton")
        hinton = ca.hinton_matrix(lca)
        print("treemap")
        treemap = self.get_treemap(gt["nodes"], gt["edges"], lca)
        print("herfindahl")
        herfindahl = herfindahl_index(lca.characterized_inventory.data)
        print("concentration")
        concentration = concentration_ratio(lca.characterized_inventory.data)
        print("MC:")
        monte_carlo = self.get_monte_carlo()

        activity_data = []
        for k, v in self.activity.items():
            obj = get_activity(k)
            activity_data.append((obj["name"], "%.2g" % v, obj["unit"]))

        self.report = {
            "activity": activity_data,
            "method": {
                "name": ": ".join(self.method),
                "unit": methods[self.method]["unit"],
            },
            "score": float(lca.score),
            "contribution": {
                "hinton": hinton,
                "treemap": treemap,
                "herfindahl": herfindahl,
                "concentration": concentration,
            },
            "force_directed": force_directed,
            "monte carlo": monte_carlo,
            "metadata": {
                "type": "Brightway2 serialized LCA report",
                "version": self.version,
                "uuid": self.uuid,
            },
        }

    def get_treemap(self, nodes, edges, lca, unroll_cutoff=0.01, simplify_limit=0.1):
        nodes, edges, links = GTManipulator.unroll_graph(
            nodes, edges, lca.score, cutoff=unroll_cutoff
        )
        nodes, edges = GTManipulator.simplify(
            nodes, edges, lca.score, limit=simplify_limit
        )
        return GTManipulator.d3_treemap(nodes, edges, lca)

    def get_monte_carlo(self):
        """Get Monte Carlo results"""
        print("Entered get_monte_carlo")
        if not self.iterations:
            # No Monte Carlo desired
            return None
        mc_data = ParallelMonteCarlo(
            self.activity, self.method, iterations=self.iterations, cpus=self.cpus
        ).calculate()
        print("Converting to array")
        mc_data = np.array(mc_data)
        print("Checking shape")
        if np.unique(mc_data).shape[0] == 1:
            # No uncertainty in database
            return None
        print("Finished MC .calculate(); Sorting")
        mc_data.sort()
        # Filter outliers
        print("Filter outliers")
        offset = int(self.outliers * mc_data.shape[0])
        lower = mc_data[offset]
        upper = mc_data[-offset]
        mc_data = mc_data[offset:-offset]
        num_bins = max(100, min(20, int(np.sqrt(self.iterations))))
        # Gaussian KDE to smooth fit
        print("KDE smoothing")
        kde = gaussian_kde(mc_data)
        kde_xs = np.linspace(mc_data.min(), mc_data.max(), 500)
        kde_ys = kde.evaluate(kde_xs)
        # Histogram
        print("Histogram")
        hist_ys, hist_xs = np.histogram(mc_data, bins=num_bins, density=True)
        hist_xs = np.repeat(hist_xs, 2)
        hist_ys = np.hstack(
            (
                np.array(0),
                np.repeat(hist_ys, 2),
                np.array(0),
            )
        )
        print("Finished .get_monte_carlo")
        return {
            "smoothed": zip(kde_xs.tolist(), kde_ys.tolist()),
            "histogram": zip(hist_xs.tolist(), hist_ys.tolist()),
            "statistics": {
                "median": float(np.median(mc_data)),
                "mean": float(np.mean(mc_data)),
                "interval": [float(lower), float(upper)],
            },
        }

    def get_force_directed(self, nodes, edges, lca):
        """Get graph traversal results"""
        nodes, edges = GTManipulator.simplify_naive(nodes, edges, lca.score)
        nodes = GTManipulator.add_metadata(nodes, lca)
        return GTManipulator.d3_force_directed(nodes, edges, lca.score)

    def write(self):
        """Write report data to file"""
        dirpath = projects.request_directory("reports")
        filepath = os.path.join(dirpath, "report.%s.json" % self.uuid)
        JsonWrapper.dump(self.report, filepath)

    def upload(self):
        """Upload report data if allowed"""
        if not config.p.get("upload_reports", False) or not config.p.get(
            "report_server_url", None
        ):
            raise ValueError("Report uploading not allowed")
        url = config.p["report_server_url"]
        if url[-1] != "/":
            url += "/"
        r = requests.post(
            url + "upload",
            data=JsonWrapper.dumps(self.report),
            headers={"content-type": "application/json"},
        )
        if r.status_code == 200:
            report_url = url + "report/" + self.uuid
            self.report["metadata"]["online"] = report_url
            return report_url
        else:
            return False



import sys
from warnings import warn

import bw2calc as bc
import numpy as np
import pyprind
from bw2data import Database, databases, get_activity, methods


def contribution_for_all_datasets_one_method(database, method, progress=True):
    """Calculate contribution analysis (for technosphere processes) for all inventory datasets in one database for one LCIA method.

    Args:
        *database* (str): Name of database
        *method* (tuple): Method tuple

    Returns:
        NumPy array of relative contributions. Each column sums to one.
        Lookup dictionary, dataset keys to row/column indices

    """

    def get_normalized_scores(lca, kind):
        if kind == "activities":
            data = lca.characterized_inventory.sum(axis=0)
        elif kind == "flows":
            data = lca.characterized_inventory.sum(axis=1)
        elif kind == "all":
            data = lca.characterized_inventory.data
        scores = np.abs(np.array(data).ravel())
        summed = scores.sum()
        if summed == 0:
            return np.zeros(scores.shape)
        else:
            return scores / summed

    assert database in databases, f"Can't find database {database}"
    assert method in methods, f"Can't find method {method}"
    db = Database(database)
    assert len(db), f"Database {database} appears to have no datasets"

    # Array to store results
    results = np.zeros((len(db), len(db)), dtype=np.float32)

    # Instantiate LCA object
    lca = bc.LCA({db.random(): 1}, method=method)
    lca.lci()
    lca.lcia()

    rows = lca.characterized_inventory.shape[0]
    cols = lca.characterized_inventory.shape[1]
    all_cutoff = cols * 4

    results = {
        "activities": np.zeros((cols, cols), dtype=np.float32),
        "flows": np.zeros((rows, cols), dtype=np.float32),
        "all": np.zeros((all_cutoff, cols), dtype=np.float32),
    }

    pbar = pyprind.ProgBar(len(db), title="Activities:")

    # Actual calculations
    for ds in db:
        lca.redo_lcia({ds.id: 1})
        if not lca.score:
            continue

        col = lca.dicts.activity[ds.id]
        results["activities"][:, col] = get_normalized_scores(lca, "activities")
        results["flows"][:, col] = get_normalized_scores(lca, "flows")
        results_all = get_normalized_scores(lca, "all")
        results_all.sort()
        results_all = results_all[::-1]
        fill_number = results_all.shape[0]
        assert fill_number < all_cutoff, "Too many values in 'all'"
        results["all"][:fill_number, col] = results_all

        pbar.update()

    print(pbar)

    return results


def print_recursive_calculation(
    activity,
    lcia_method,
    amount=1,
    max_level=3,
    cutoff=1e-2,
    file_obj=None,
    tab_character="  ",
    level=0,
    lca_obj=None,
    total_score=None,
    first=True,
):
    """Traverse a supply chain graph, and calculate the LCA scores of each component. Prints the result with the format:

    {tab_character * level }{fraction of total score} ({absolute LCA score for this input} | {amount of input}) {input activity}

    Args:
        activity: ``Activity``. The starting point of the supply chain graph.
        lcia_method: tuple. LCIA method to use when traversing supply chain graph.
        amount: int. Amount of ``activity`` to assess.
        max_level: int. Maximum depth to traverse.
        cutoff: float. Fraction of total score to use as cutoff when deciding whether to traverse deeper.
        file_obj: File-like object (supports ``.write``), optional. Output will be written to this object if provided.
        tab_character: str. Character to use to indicate indentation.

    Internal args (used during recursion, do not touch);
        level: int.
        lca_obj: ``LCA``.
        total_score: float.
        first: bool.

    Returns:
        Nothing. Prints to ``sys.stdout`` or ``file_obj``

    """
    activity = get_activity(activity)
    if file_obj is None:
        file_obj = sys.stdout

    if lca_obj is None:
        lca_obj = bc.LCA({activity: amount}, lcia_method)
        lca_obj.lci()
        lca_obj.lcia()
        total_score = lca_obj.score
    elif total_score is None:
        raise ValueError
    else:
        lca_obj.redo_lcia({activity.id: amount})
        if abs(lca_obj.score) <= abs(total_score * cutoff):
            return
    if first:
        file_obj.write("Fraction of score | Absolute score | Amount | Activity\n")
    message = "{}{:04.3g} | {:5.4n} | {:5.4n} | {:.70}".format(
        tab_character * level,
        lca_obj.score / total_score,
        lca_obj.score,
        float(amount),
        str(activity),
    )
    file_obj.write(message + "\n")
    if level < max_level:
        prod_exchanges = list(activity.production())
        if not prod_exchanges:
            prod_amount = 1
        elif len(prod_exchanges) > 1:
            warn("Hit multiple production exchanges; aborting in this branch")
            return
        else:
            prod_amount = lca_obj.technosphere_matrix[
                lca_obj.dicts.product[prod_exchanges[0].input.id],
                lca_obj.dicts.activity[prod_exchanges[0].output.id],
            ]

        for exc in activity.technosphere():
            if exc.input.id == exc.output.id:
                continue
            print_recursive_calculation(
                activity=exc.input,
                lcia_method=lcia_method,
                amount=amount * exc["amount"] / prod_amount,
                max_level=max_level,
                cutoff=cutoff,
                first=False,
                file_obj=file_obj,
                tab_character=tab_character,
                lca_obj=lca_obj,
                total_score=total_score,
                level=level + 1,
            )


def print_recursive_supply_chain(
    activity,
    amount=1,
    max_level=2,
    cutoff=0,
    file_obj=None,
    tab_character="  ",
    level=0,
):
    """Traverse a supply chain graph, and prints the inputs of each component.

    This function is only for exploration; use ``bw2calc.GraphTraversal`` for a better performing function.

    The results displayed here can also be incorrect if

    Args:
        activity: ``Activity``. The starting point of the supply chain graph.
        amount: int. Supply chain inputs will be scaled to this value.
        max_level: int. Max depth to search for.
        cutoff: float. Inputs with amounts less than ``amount * cutoff`` will not be printed or traversed further.
        file_obj: File-like object (supports ``.write``), optional. Output will be written to this object if provided.
        tab_character: str. Character to use to indicate indentation.
        level: int. Current level of the calculation. Only used internally, do not touch.

    Returns:
        Nothing. Prints to ``stdout`` or ``file_obj``

    """
    activity = get_activity(activity)
    if file_obj is None:
        file_obj = sys.stdout

    if cutoff > 0 and amount < cutoff:
        return
    message = "{}{:.3g}: {:.70}".format(tab_character * level, amount, str(activity))
    file_obj.write(message + "\n")
    if level < max_level:
        prod_exchanges = list(activity.production())
        if not prod_exchanges:
            prod_amount = 1
        elif len(prod_exchanges) > 1:
            warn("Hit multiple production exchanges; aborting in this branch")
            return
        else:
            prod_amount = prod_exchanges[0]["amount"]
            for other in activity.technosphere():
                if other.input == prod_exchanges[0].input:
                    prod_amount -= other["amount"]

        for exc in activity.technosphere():
            if exc.input.id == exc.output.id:
                continue
            print_recursive_supply_chain(
                activity=exc.input,
                amount=amount * exc["amount"] / prod_amount,
                level=level + 1,
                max_level=max_level,
                cutoff=cutoff,
                file_obj=file_obj,
                tab_character=tab_character,
            )



import pandas as pd
from bw2data import get_activity
from bw2calc import LCA


def get_labeled_inventory(lca: LCA) -> pd.DataFrame:
    """
    Take an LCA's inventory matrix and labels its rows (biosphere) and columns (technosphere) with activity metadata.

    Args:
        * *lca* (bw2calc.LCA): LCA object whose life cycle inventory has been calculated previously.

    Returns:
        pd.DataFrame with activity information as row and column MultiIndices.
    """

    assert hasattr(
        lca, "inventory"
    ), "Must calculate life cycle inventory first. Please call lci()."

    rows = [
        get_activity(lca.dicts.biosphere.reversed[i]).as_dict()
        for i in range(lca.inventory.shape[0])
    ]
    columns = [
        get_activity(lca.dicts.activity.reversed[i]).as_dict()
        for i in range(lca.inventory.shape[1])
    ]

    return pd.DataFrame(
        data=lca.inventory.todense(),
        index=pd.MultiIndex.from_frame(pd.DataFrame(rows)),
        columns=pd.MultiIndex.from_frame(pd.DataFrame(columns)),
    )



import copy
import itertools
from heapq import heappop, heappush

from bw2data import config, get_activity


def tupify(o):
    """Transform edge from dict to tuples. Multiply impact by -1 because sort by min, not max"""
    return (-1 * o["impact"], o["from"], o["to"], o["amount"], o["exc_amount"])


class GTManipulator:
    """Manipulate ``GraphTraversal`` results."""

    @staticmethod
    def unroll_graph(nodes, edges, score, cutoff=0.005, max_links=2500):
        """Unroll a ``GraphTraversal`` result, allowing the same activity to appear in the graph multiple times."""
        input_nodes, input_edges = copy.deepcopy(nodes), copy.deepcopy(edges)
        counter = itertools.count()
        edges_dict = {}
        heap, edges, links, cutoff_score = [], [], 0, cutoff * score
        nodes = {-1: copy.deepcopy(input_nodes[-1])}

        for edge in input_edges:
            edges_dict.setdefault(edge["to"], []).append(edge)

        for edge in edges_dict[-1]:
            heappush(heap, tupify(edge))
            links += 1

        while heap and links < max_links:
            score, from_, to, cum_amount, exc_amount = heappop(heap)

            if (-1 * score) < cutoff_score:
                continue

            # ``from_`` is old-style indexing, ``to`` is new-style indexing
            from_node = input_nodes[from_]
            to_node = nodes[to]
            new_amount = exc_amount * to_node["amount"]

            node_id = next(counter)
            # Only node that doesn't have a ``row`` attribute is the
            # functional unit, which by definition has no outgoing links
            nodes[node_id] = {
                "amount": new_amount,
                "row": from_,
                "cum": from_node["cum"] * new_amount / from_node["amount"],
                "ind": from_node["ind"],
            }
            edges.append(
                {
                    "to": to,
                    "from": node_id,
                    "amount": new_amount,
                    "exc_amount": exc_amount,
                    "impact": nodes[node_id]["cum"],
                }
            )

            frommer = lambda x: x["from"]
            # Include all edges for now; ignore minor ones when popping heap
            for edge in sorted(edges_dict.get(from_, []), key=frommer):
                from_edge_node = input_nodes[from_]
                heappush(
                    heap,
                    tupify(
                        {
                            "amount": new_amount * edge["exc_amount"],
                            "from": edge["from"],
                            "to": node_id,
                            "exc_amount": edge["exc_amount"],
                            "impact": from_edge_node["cum"]
                            / from_edge_node["amount"]
                            * new_amount
                            * edge["exc_amount"],
                        }
                    ),
                )
                links += 1
        return nodes, edges, links

    @staticmethod
    def add_metadata(nodes, lca):
        """Add metadata to nodes, like name, unit, and category."""
        new_nodes = {}
        for key, value in nodes.items():
            new_value = copy.deepcopy(value)
            if key == -1:
                new_value.update(
                    {
                        "name": "Functional unit",
                        "unit": "unit",
                        "categories": ["Functional unit"],
                    }
                )
            else:
                if "row" in value:
                    index = value["row"]
                else:
                    index = key
                code = lca.dicts.activity.reversed[index]
                ds = get_activity(code)
                new_value.update(
                    {
                        "name": ds.get("name", "Unknown"),
                        "categories": ds.get("categories", []),
                        "unit": ds.get("unit", "Unknown"),
                        "key": code,
                    }
                )
            new_nodes[key] = new_value
        return new_nodes

    @staticmethod
    def d3_force_directed(nodes, edges, score):
        """Reformat to D3 style, which is a list of nodes, and edge ids are node list indices."""
        # Sort node ids by highest cumulative score first
        node_ids = [x[1] for x in sorted([(v["cum"], k) for k, v in nodes.items()])]
        new_nodes = [nodes[i] for i in node_ids]
        lookup = dict([(key, index) for index, key in enumerate(node_ids)])
        new_edges = [
            {
                "source": lookup[e["to"]],
                "target": lookup[e["from"]],
                "amount": e["impact"],
            }
            for e in edges
        ]
        return {"edges": new_edges, "nodes": new_nodes, "total": score}

    @staticmethod
    def simplify(nodes, edges, score, limit=0.005):
        """Simplify supply chain to include only nodes which individually contribute ``limit * score``.

        Only removes and combines edges; doesn't check to make sure amounts add up correctly."""
        if isinstance(limit, int) and limit > 1:
            nodes_to_delete = sorted(
                [(value["amount"] * value["ind"], key) for key, value in nodes.items()],
                reverse=True,
            )[limit:]
        else:
            nodes_to_delete = [
                key
                for key, value in nodes.items()
                if key != -1 and (value["amount"] * value["ind"]) < (score * limit)
            ]
        edges_dict = {(edge["to"], edge["from"]): edge for edge in copy.deepcopy(edges)}
        for node in nodes_to_delete:
            p_edges = [key for key in edges_dict if key[1] == node]
            c_edges = [key for key in edges_dict if key[0] == node]
            total_amount = sum([edges_dict[k]["amount"] for k in p_edges])
            total_impact = sum([edges_dict[k]["impact"] for k in p_edges])
            for p_key, c_key in itertools.product(p_edges, c_edges):
                key = (p_key[0], c_key[1])
                if p_key[0] == c_key[1]:
                    continue
                p_edge = edges_dict[p_key]
                c_edge = edges_dict[c_key]
                e_amount = c_edge["amount"] * p_edge["amount"] / total_amount
                e_impact = c_edge["impact"] * p_edge["impact"] / total_impact
                e_eamount = p_edge["exc_amount"] * c_edge["exc_amount"]
                if key in edges_dict:
                    edges_dict[key]["amount"] += e_amount
                    edges_dict[key]["exc_amount"] += e_eamount
                    edges_dict[key]["impact"] += e_impact
                else:
                    edges_dict[key] = {
                        "to": p_edge["to"],
                        "from": c_edge["from"],
                        "amount": e_amount,
                        "exc_amount": e_eamount,
                        "impact": e_impact,
                    }
            # Remove obsolete edges
            for key in p_edges:
                del edges_dict[key]
            for key in c_edges:
                del edges_dict[key]
        nodes = {
            key: value
            for key, value in nodes.items()
            if key not in set(nodes_to_delete)
        }
        return nodes, edges_dict.values()

    @staticmethod
    def simplify_naive(nodes, edges, score, limit=0.0025):
        """Naive simplification which simplifies removes links below an LCA score cutoff. Orphan nodes are also deleted."""
        edges = [e for e in edges if e["impact"] >= (score * limit)]
        good_nodes = set([e["from"] for e in edges]).union(
            set([e["to"] for e in edges])
        )
        nodes = dict([(k, v) for k, v in nodes.items() if k in good_nodes])
        return nodes, edges

    @staticmethod
    def d3_treemap(nodes, edges, lca, add_biosphere=False):
        """Add node data by traversing the graph; assign different metadata to leaf nodes."""
        ra, rp, rb = lca.reverse_dict()

        child_nodes = lambda node, edges: [e["from"] for e in edges if e["to"] == node]

        counter = itertools.count(1)

        def format_node(node_key, node_data, ra):
            if node_key == -1:
                return {
                    "name": "Functional unit",
                    "unit": "unit",
                    "location": config.global_location,
                    "categories": "",
                    "id": next(counter),
                    "amount": 1.0,
                }

            if "row" in node_data:
                node_key = node_data["row"]
            key = ra[node_key]
            ds = get_activity(key)
            return {
                "name": ds.get("name", "Unknown"),
                "unit": ds.get("unit", "Unknown"),
                "location": ds.get("location", "Unknown"),
                "categories": ", ".join(ds.get("categories", [])),
                "id": next(counter),
                "amount": node_data["amount"],
            }

        def format_child_node(node_key, node_data, ra, add_biosphere):
            ds = format_node(node_key, node_data, ra)
            ds["size"] = node_data["cum"]
            ds["variance"] = 0.5
            return ds

        def process_node(node):
            cn = child_nodes(node, edges)
            if cn:
                ds = format_node(node, nodes[node], ra)
                ds["children"] = [process_node(o) for o in cn]
                return ds
            else:
                return format_child_node(node, nodes[node], ra, False)

        return process_node(-1)



import numpy as np
from bw2data import get_activity


class ContributionAnalysis:
    def sort_array(self, data, limit=25, limit_type="number", total=None):
        """
        Common sorting function for all ``top`` methods. Sorts by highest value first.

        Operates in either ``number`` or ``percent`` mode. In ``number`` mode, return ``limit`` values. In ``percent`` mode, return all values >= (total * limit); where ``0 < limit <= 1``.

        Returns 2-d numpy array of sorted values and row indices, e.g.:

        .. code-block:: python

            ContributionAnalysis().sort_array((1., 3., 2.))

        returns

        .. code-block:: python

            (
                (3, 1),
                (2, 2),
                (1, 0)
            )

        Args:
            * *data* (numpy array): A 1-d array of values to sort.
            * *limit* (number, default=25): Number of values to return, or percentage cutoff.
            * *limit_type* (str, default=``number``): Either ``number`` or ``percent``.
            * *total* (number, default=None): Optional specification of summed data total.

        Returns:
            2-d numpy array of values and row indices.

        """
        total = total or np.abs(data).sum()
        if limit_type not in ("number", "percent"):
            raise ValueError("limit_type must be either 'percent' or 'index'.")
        if limit_type == "percent":
            if not 0 < limit <= 1:
                raise ValueError("Percentage limits > 0 and <= 1.")
            limit = (np.abs(data) >= (total * limit)).sum()

        results = np.hstack(
            (data.reshape((-1, 1)), np.arange(data.shape[0]).reshape((-1, 1)))
        )
        return results[np.argsort(np.abs(data))[::-1]][:limit, :]

    def top_matrix(self, matrix, rows=5, cols=5):
        """
        Find most important (i.e. highest summed) rows and columns in a matrix, as well as the most corresponding non-zero individual elements in the top rows and columns.

        Only returns matrix values which are in the top rows and columns. Element values are returned as a tuple: ``(row, col, row index in top rows, col index in top cols, value)``.

        Example:

        .. code-block:: python

            matrix = [
                [0, 0, 1, 0],
                [2, 0, 4, 0],
                [3, 0, 1, 1],
                [0, 7, 0, 1],
            ]

        In this matrix, the row sums are ``(1, 6, 5, 8)``, and the columns sums are ``(5, 7, 6, 2)``. Therefore, the top rows are ``(3, 1)`` and the top columns are ``(1, 2)``. The result would therefore be:

        .. code-block:: python

            (
                (
                    (3, 1, 0, 0, 7),
                    (3, 2, 0, 1, 1),
                    (1, 2, 1, 1, 4)
                ),
                (3, 1),
                (1, 2)
            )

        Args:
            * *matrix* (array or matrix): Any Python object that supports the ``.sum(axis=)`` syntax.
            * *rows* (int): Number of rows to select.
            * *cols* (int): Number of columns to select.

        Returns:
            (elements, top rows, top columns)
        """
        top_rows = np.argsort(np.abs(np.array(matrix.sum(axis=1)).ravel()))[
            : -rows - 1 : -1
        ]
        top_cols = np.argsort(np.abs(np.array(matrix.sum(axis=0)).ravel()))[
            : -cols - 1 : -1
        ]
        elements = []
        for row, x in enumerate(top_rows):
            for col, y in enumerate(top_cols):
                if matrix[x, y] != 0:
                    elements.append((x, y, row, col, float(matrix[x, y])))
        return elements, top_rows.astype(int), top_cols.astype(int)

    def hinton_matrix(self, lca, rows=5, cols=5):
        coo, b, t = self.top_matrix(lca.characterized_inventory, rows=rows, cols=cols)
        coo = [row[2:] for row in coo]  # Don't need matrix indices
        flows = [self.get_name(lca.dicts.biosphere.reversed[x]) for x in b]
        activities = [self.get_name(lca.dicts.activity.reversed[x]) for x in t]
        return {
            "results": coo,
            "total": lca.score,
            "xlabels": activities,
            "ylabels": flows,
        }

    def annotate(self, sorted_data, rev_mapping):
        """Reverse the mapping from database ids to array indices"""
        return [(row[0], rev_mapping[row[1]]) for row in sorted_data]

    def top_processes(self, matrix, **kwargs):
        """Return an array of [value, index] technosphere processes."""
        return self.sort_array(np.array(matrix.sum(axis=0)).ravel(), **kwargs)

    def top_emissions(self, matrix, **kwargs):
        """Return an array of [value, index] biosphere emissions."""
        return self.sort_array(np.array(matrix.sum(axis=1)).ravel(), **kwargs)

    def annotated_top_processes(self, lca, names=True, **kwargs):
        """Get list of most damaging processes in an LCA, sorted by ``abs(direct impact)``.

        Returns a list of tuples: ``(lca score, supply, activity)``. If ``names`` is False, they returns the process key as the last element.

        """
        results = [
            (
                score,
                lca.supply_array[int(index)],
                lca.dicts.activity.reversed[int(index)],
            )
            for score, index in self.top_processes(
                lca.characterized_inventory, **kwargs
            )
        ]
        if names:
            results = [(x[0], x[1], get_activity(x[2])) for x in results]
        return results

    def annotated_top_emissions(self, lca, names=True, **kwargs):
        """Get list of most damaging biosphere flows in an LCA, sorted by ``abs(direct impact)``.

        Returns a list of tuples: ``(lca score, inventory amount, activity)``. If ``names`` is False, they returns the process key as the last element.

        """
        results = [
            (score, lca.inventory[index, :].sum(), lca.dicts.biosphere.reversed[index])
            for score, index in self.top_emissions(
                lca.characterized_inventory, **kwargs
            )
        ]
        if names:
            results = [(x[0], x[1], get_activity(x[2])) for x in results]
        return results

    def get_name(self, key):
        return get_activity(key).get("name", "Unknown")

    def d3_treemap(
        self, matrix, rev_bio, rev_techno, limit=0.025, limit_type="percent"
    ):
        """
        Construct treemap input data structure for LCA result. Output like:

        .. code-block:: python

            {
            "name": "LCA result",
            "children": [{
                "name": process 1,
                "children": [
                    {"name": emission 1, "size": score},
                    {"name": emission 2, "size": score},
                    ],
                }]
            }

        """
        total = np.abs(matrix).sum()
        processes = self.top_processes(matrix, limit=limit, limit_type=limit_type)
        data = {"name": "LCA result", "children": [], "size": total}
        for _, tech_index in processes:
            name = self.get_name(rev_techno[tech_index])
            this_score = np.abs(matrix[:, int(tech_index)].toarray().ravel()).sum()
            children = []
            for score, bio_index in self.sort_array(
                matrix[:, int(tech_index)].toarray().ravel(),
                limit=limit,
                limit_type=limit_type,
                total=total,
            ):
                children.append(
                    {
                        "name": self.get_name(rev_bio[bio_index]),
                        "size": float(abs(matrix[int(bio_index), int(tech_index)])),
                    }
                )
            children_score = sum([x["size"] for x in children])
            if children_score < (0.95 * this_score):
                children.append({"name": "Others", "size": this_score - children_score})
            data["children"].append(
                {
                    "name": name,
                    "size": this_score,
                    # "children": children
                }
            )
        return data



import math
import operator
from os.path import commonprefix

import bw2calc as bc
import bw2data as bd
import numpy as np
import pandas as pd
import tabulate
from pandas import DataFrame


def aggregated_dict(activity):
    """Return dictionary of inputs aggregated by input reference product."""
    results = {}
    for exc in activity.technosphere():
        results[exc.input["reference product"]] = (
            results.setdefault(exc.input["reference product"], 0) + exc["amount"]
        )

    for exc in activity.biosphere():
        results[exc.input["name"]] = (
            results.setdefault(exc.input["name"], 0) + exc["amount"]
        )

    return results


def compare_dictionaries(one, two, rel_tol=1e-4, abs_tol=1e-9):
    """Compare two dictionaries with form ``{str: float}``, and return a set of keys where differences where present.

    Tolerance values are inputs to `math.isclose <https://docs.python.org/3/library/math.html#math.isclose>`__."""
    return (
        set(one)
        .symmetric_difference(set(two))
        .union(
            {
                key
                for key in one
                if key in two
                and not math.isclose(
                    a=one[key], b=two[key], rel_tol=rel_tol, abs_tol=abs_tol
                )
            }
        )
    )


def find_differences_in_inputs(
    activity, rel_tol=1e-4, abs_tol=1e-9, locations=None, as_dataframe=False
):
    """Given an ``Activity``, try to see if other activities in the same database (with the same name and
    reference product) have the same input levels.

    Tolerance values are inputs to `math.isclose <https://docs.python.org/3/library/math.html#math.isclose>`__.

    If differences are present, a difference dictionary is constructed, with the form:

    .. code-block:: python

        {Activity instance: [(name of input flow (str), amount)]}

    Note that this doesn't reference a specific exchange, but rather sums **all exchanges with the same input reference product**.

    Assumes that all similar activities produce the same amount of reference product.

    ``(x, y)``, where ``x`` is the number of similar activities, and ``y`` is a dictionary of the differences. This dictionary is empty if no differences are found.

    Args:
        activity: ``Activity``. Activity to analyze.
        rel_tol: float. Relative tolerance to decide if two inputs are the same. See above.
        abs_tol: float. Absolute tolerance to decide if two inputs are the same. See above.
        locations: list, optional. Locations to restrict comparison to, if present.
        as_dataframe: bool. Return results as pandas DataFrame.

    Returns:
        dict or ``pandas.DataFrame``.


    """
    assert isinstance(activity, bd.backends.proxies.Activity)

    try:
        similar = [
            obj
            for obj in bd.Database(activity["database"])
            if obj != activity
            and obj.get("reference product") == activity.get("reference product")
            and obj.get("name") == activity["name"]
            and (not locations or obj.get("location") in locations)
        ]
    except KeyError:
        raise ValueError("Given activity has no `name`; can't find similar names")

    result = {}

    origin_dict = aggregated_dict(activity)

    for target in similar:
        target_dict = aggregated_dict(target)
        difference = compare_dictionaries(origin_dict, target_dict, rel_tol, abs_tol)
        if difference:
            if activity not in result:
                result[activity] = {}
            result[activity].update(
                {key: value for key, value in origin_dict.items() if key in difference}
            )
            result[target] = {
                key: value for key, value in target_dict.items() if key in difference
            }

    if as_dataframe:
        df = DataFrame(
            [{"location": obj.get("location"), **result[obj]} for obj in result]
        )
        df.set_index("location", inplace=True)
        return df
    else:
        return result


def compare_activities_by_lcia_score(activities, lcia_method, band=0.1):
    """Compare selected activities to see if they are substantially different.

    Substantially different means that all LCIA scores lie within a band of ``band * max_lcia_score``.

    Inputs:

        ``activities``: List of ``Activity`` objects.
        ``lcia_method``: Tuple identifying a ``Method``

    Returns:

        Nothing, but prints to stdout.

    """
    import bw2calc as bc

    activities = [bd.get_activity(obj) for obj in activities]

    lca = bc.LCA({a: 1 for a in activities}, lcia_method)
    lca.lci()
    lca.lcia()

    # First pass: Are all scores close?
    scores = []

    for a in activities:
        lca.redo_lcia({a.id: 1})
        scores.append(lca.score)

    if abs(max(scores) - min(scores)) < band * abs(max(scores)):
        print("All activities similar")
        return
    else:
        print("Differences observed. LCA scores:")
        for x, y in zip(scores, activities):
            print("\t{:5.3f} -> {}".format(x, y.key))


def find_leaves(
    activity,
    lcia_method,
    results=None,
    lca_obj=None,
    amount=1,
    total_score=None,
    level=0,
    max_level=3,
    cutoff=2.5e-2,
):
    """Traverse the supply chain of an activity to find leaves - places where the impact of that
    component falls below a threshold value.

    Returns a list of ``(impact of this activity, amount consumed, Activity instance)`` tuples."""
    first_level = results is None

    activity = bd.get_activity(activity)

    if first_level:
        level = 0
        results = []

        lca_obj = bc.LCA({activity: amount}, lcia_method)
        lca_obj.lci()
        lca_obj.lcia()
        total_score = lca_obj.score
    else:
        lca_obj.redo_lcia({activity.id: amount})

        # If this is a leaf, add the leaf and return
        if abs(lca_obj.score) <= abs(total_score * cutoff) or level >= max_level:

            # Only add leaves with scores that matter
            if abs(lca_obj.score) > abs(total_score * 1e-4):
                results.append((lca_obj.score, amount, activity))
            return results

        else:
            # Add direct emissions from this demand
            direct = (
                lca_obj.characterization_matrix
                * lca_obj.biosphere_matrix
                * lca_obj.demand_array
            ).sum()
            if abs(direct) >= abs(total_score * 1e-4):
                results.append((direct, amount, activity))

    for exc in activity.technosphere():
        find_leaves(
            activity=exc.input,
            lcia_method=lcia_method,
            results=results,
            lca_obj=lca_obj,
            amount=amount * exc["amount"],
            total_score=total_score,
            level=level + 1,
            max_level=max_level,
            cutoff=cutoff,
        )

    return sorted(results, reverse=True)


def get_cpc(activity):
    try:
        return next(
            cl[1] for cl in activity.get("classifications", []) if cl[0] == "CPC"
        )
    except StopIteration:
        return


def get_value_for_cpc(lst, label):
    for elem in lst:
        if elem[2] == label:
            return elem[0]
    return 0


def group_leaves(leaves):
    """Group elements in ``leaves`` by their `CPC (Central Product Classification) <https://unstats.un.org/unsd/classifications/Econ/cpc>`__ code.

    Returns a list of ``(fraction of total impact, specific impact, amount, Activity instance)`` tuples."""
    results = {}

    for leaf in leaves:
        cpc = get_cpc(leaf[2])
        if cpc not in results:
            results[cpc] = np.zeros((2,))
        results[cpc] += np.array(leaf[:2])

    return sorted([v.tolist() + [k] for k, v in results.items()], reverse=True)


def compare_activities_by_grouped_leaves(
    activities,
    lcia_method,
    mode="relative",
    max_level=4,
    cutoff=7.5e-3,
    output_format="list",
    str_length=50,
):
    """Compare activities by the impact of their different inputs, aggregated by the product classification of those inputs.

    Args:
        activities: list of ``Activity`` instances.
        lcia_method: tuple. LCIA method to use when traversing supply chain graph.
        mode: str. If "relative" (default), results are returned as a fraction of total input. Otherwise, results are absolute impact per input exchange.
        max_level: int. Maximum level in supply chain to examine.
        cutoff: float. Fraction of total impact to cutoff supply chain graph traversal at.
        output_format: str. See below.
        str_length; int. If ``output_format`` is ``html``, this controls how many characters each column label can have.

    Raises:
        ValueError: ``activities`` is malformed.

    Returns:
        Depends on ``output_format``:

        * ``list``: Tuple of ``(column labels, data)``
        * ``html``: HTML string that will print nicely in Jupyter notebooks.
        * ``pandas``: a pandas ``DataFrame``.

    """
    for act in activities:
        if not isinstance(act, bd.backends.proxies.Activity):
            raise ValueError("`activities` must be an iterable of `Activity` instances")

    objs = [
        group_leaves(find_leaves(act, lcia_method, max_level=max_level, cutoff=cutoff))
        for act in activities
    ]
    sorted_keys = sorted(
        [
            (max([el[0] for obj in objs for el in obj if el[2] == key]), key)
            for key in {el[2] for obj in objs for el in obj}
        ],
        reverse=True,
    )
    name_common = commonprefix([act["name"] for act in activities])

    if " " not in name_common:
        name_common = ""
    else:
        last_space = len(name_common) - operator.indexOf(reversed(name_common), " ")
        name_common = name_common[:last_space]
        print("Omitting activity name common prefix: '{}'".format(name_common))

    product_common = commonprefix(
        [act.get("reference product", "") for act in activities]
    )

    lca = bc.LCA({act: 1 for act in activities}, lcia_method)
    lca.lci()
    lca.lcia()

    labels = [
        "activity",
        "product",
        "location",
        "unit",
        "total",
        "direct emissions",
    ] + [key for _, key in sorted_keys]
    data = []
    for act, lst in zip(activities, objs):
        lca.redo_lcia({act.id: 1})
        data.append(
            [
                act["name"].replace(name_common, ""),
                act.get("reference product", "").replace(product_common, ""),
                act.get("location", "")[:25],
                act.get("unit", ""),
                lca.score,
            ]
            + [
                (
                    lca.characterization_matrix
                    * lca.biosphere_matrix
                    * lca.demand_array
                ).sum()
            ]
            + [get_value_for_cpc(lst, key) for _, key in sorted_keys]
        )

    data.sort(key=lambda x: x[4], reverse=True)

    if mode == "relative":
        for row in data:
            for index, point in enumerate(row[5:]):
                row[index + 5] = point / row[4]

    if output_format == "list":
        return labels, data
    elif output_format == "pandas":
        return pd.DataFrame(data, columns=labels)
    elif output_format == "html":
        return tabulate.tabulate(
            data,
            [x[:str_length] for x in labels],
            tablefmt="html",
            floatfmt=".3f",
        )



from setuptools import setup


v_temp = {}
with open("bw2analyzer/version.py") as fp:
    exec(fp.read(), v_temp)
version = ".".join((str(x) for x in v_temp["version"]))


setup(
    name="bw2analyzer",
    version=version,
    packages=["bw2analyzer"],
    author="Chris Mutel",
    author_email="cmutel@gmail.com",
    license=open("LICENSE.txt").read(),
    install_requires=[
        "bw2calc",
        "bw2data",
        "matplotlib",
        "numpy",
        "pandas",
        "pyprind",
        "requests",
        "scipy",
        "stats_arrays",
        "tabulate",
    ],
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/cmutel/brightway2-analyzer",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
)



from collections import defaultdict
from warnings import warn

from bw2calc import LCA
from bw2data import Method, get_activity, Database


def traverse_tagged_databases(
    functional_unit, method, label="tag", default_tag="other", secondary_tags=[], fg_databases=None
):

    """Traverse a functional unit throughout its foreground database(s) or the 
    listed databses in fg_databses, and group impacts by tag label.

    Contribution analysis work by linking impacts to individual activities.
    However, you also might want to group impacts in other ways. For example,
    give individual biosphere exchanges their own grouping, or aggregate two
    activities together.

    Consider this example system, where the letters are the tag labels, and the
    numbers are exchange amounts. The functional unit is one unit of the tree
    root.

    .. image:: images/tagged-traversal.png
       :alt: Example tagged supply chain

    In this supply chain, tags are applied to activities and biosphere exchanges.
    If a biosphere exchange is not tagged, it inherits the tag of its producing
    activity. Similarly, links to other databases are assessed with the usual
    LCA machinery, and the total LCA score is tagged according to its consuming
    activity. If an activity does not have a tag, a default tag is applied.

    We can change our visualization to show the use of the default tags:

    .. image:: images/tagged-traversal-2.png
       :alt: Example tagged supply chain

    And then we can manually calculate the tagged impacts. Normally we would
    need to know the actual biosphere flows and their respective
    characterization factors (CF), but in this example we assume that each
    CF is one. Our result, group by tags, would therefore be:

        * **A**: :math:`6 + 27 = 33`
        * **B**: :math:`30 + 44 = 74`
        * **C**: :math:`5 + 16 + 48 = 69`
        * **D**: :math:`14`

    This function will only traverse the foreground database, i.e. the
    database of the functional unit activity. A functional unit can have
    multiple starting nodes; in this case, all foreground databases are
    traversed.

    Input arguments:

        * ``functional_unit``: A functional unit dictionary, e.g. ``{("foo", "bar"): 42}``.
        * ``method``: A method name, e.g. ``("foo", "bar")``
        * ``label``: The label of the tag classifier. Default is ``"tag"``
        * ``default_tag``: The tag classifier to use if none was given. Default is ``"other"``
        * ``secondary_tags``: List of tuples in the format (secondary_label, secondary_default_tag). Default is empty list.
        * ``fg_databases``: a list of foreground databases to be traversed, e.g. ['foreground', 'biomass', 'machinery']
                            It's not recommended to include all databases of a project in the list to be traversed, especially not ecoinvent itself

    Returns:

        Aggregated tags dictionary from ``aggregate_tagged_graph``, and tagged supply chain graph from ``recurse_tagged_database``.

    """

    lca = LCA(functional_unit, method)
    lca.lci()
    lca.lcia()

    method_dict = {o[0]: o[1] for o in Method(method).load()}

    graph = [
        recurse_tagged_database(
            key, amount, method_dict, lca, label, default_tag, secondary_tags, fg_databases
        )
        for key, amount in functional_unit.items()
    ]

    return aggregate_tagged_graph(graph), graph


def aggregate_tagged_graph(graph):
    """Aggregate a graph produced by ``recurse_tagged_database`` by the provided tags.

    Outputs a dictionary with keys of tags and numeric values.

    .. code-block:: python

        {'a tag': summed LCIA scores}

    """

    def recursor(obj, scores):
        scores[obj["tag"]] += obj["impact"]
        for flow in obj["biosphere"]:
            scores[flow["tag"]] += flow["impact"]
        for exc in obj["technosphere"]:
            scores = recursor(exc, scores)
        return scores

    scores = defaultdict(int)
    for obj in graph:
        scores = recursor(obj, scores)
    return scores


def recurse_tagged_database(
    activity, amount, method_dict, lca, label, default_tag, secondary_tags=[], fg_databases=None, warned=False
):

    """Traverse a foreground database and assess activities and biosphere flows by tags.


    Input arguments:

        * ``activity``: Activity tuple or object
        * ``amount``: float
        * ``method_dict``: Dictionary of biosphere flow tuples to CFs, e.g. ``{("biosphere", "foo"): 3}``
        * ``lca``: An ``LCA`` object that is already initialized, i.e. has already calculated LCI and LCIA with same method as in ``method_dict``
        * ``label``: string
        * ``default_tag``: string
        * ``secondary_tags``: List of tuples in the format (secondary_label, secondary_default_tag). Default is empty list.
        
        * ``fg_databases``: a list of foreground databases to be traversed, e.g. ['foreground', 'biomass', 'machinery']
                            It's not recommended to include all databases of a project in the list to be traversed, especially not ecoinvent itself

    Returns:

    .. code-block:: python

        {
            'activity': activity object,
            'amount': float,
            'tag': string,
            'secondary_tags': [list of strings],
            'impact': float (impact of inputs from outside foreground database),
            'biosphere': [{
                'amount': float,
                'impact': float,
                'tag': string,
                'secondary_tags': [list of strings]
            }],
            'technosphere': [this data structure]
        }

    """
    if isinstance(activity, tuple):
        activity = get_activity(activity)

    MESSAGE = """Given databases include many activities, and traversal may be slow.
Consider using `GraphTraversalLCA` from `bw2calc` instead."""

    if fg_databases is None:  # then set the list equal to the database of the functional unit
        fg_databases = [activity['database']]  # list, single item
    elif not warned and sum(len(Database(name)) for name in fg_databases) > 2500:
        warn(MESSAGE)
        warned = True

    inputs = list(activity.technosphere())
    production = list(activity.production())

    if not production:
        scale = 1
    elif len(production) > 1:
        warn("Hit multiple production exchanges; aborting in this branch")
        return
    else:
        scale = production[0]["amount"]
        for other in activity.technosphere():
            if other.input == production[0].input:
                scale -= other["amount"]

    inside = [exc for exc in inputs if exc.input["database"] in fg_databases]

    outside = {
        exc.input.id: exc["amount"] / scale * amount
        for exc in inputs
        if exc["input"][0] not in fg_databases
    }

    if outside:
        lca.redo_lcia(outside)
        outside_score = lca.score
    else:
        outside_score = 0

    return {
        "activity": activity,
        "amount": amount,
        "tag": activity.get(label) or default_tag,
        "secondary_tags": [activity.get(t[0]) or t[1] for t in secondary_tags],
        "impact": outside_score,
        "biosphere": [
            {
                "activity": exc.input,
                "amount": exc["amount"] / scale * amount,
                "impact": exc["amount"]
                / scale
                * amount
                * method_dict.get(exc["input"], 0),
                "tag": exc.get(label) or activity.get(label) or default_tag,
                "secondary_tags": [
                    exc.get(t[0]) or activity.get(t[0]) or t[1] for t in secondary_tags
                ],
            }
            for exc in activity.biosphere()
        ],
        "technosphere": [
            recurse_tagged_database(
                activity=exc.input,
                amount=exc["amount"] / scale * amount,
                method_dict=method_dict,
                lca=lca,
                label=label,
                default_tag=default_tag,
                secondary_tags=secondary_tags,
                fg_databases=fg_databases,
                warned=warned,
            )
            for exc in inside
        ],
    }


## tagged graph functions using multiple methods


def multi_traverse_tagged_databases(
    functional_unit, methods, label="tag", default_tag="other", secondary_tags=[]
):

    """Traverse a functional unit throughout its foreground database(s), and
    group impacts (for multiple methods) by tag label.

    Input arguments:
        * ``functional_unit``: A functional unit dictionary, e.g. ``{("foo", "bar"): 42}``.
        * ``methods``: A list of method names, e.g. ``[("foo", "bar"), ("baz", "qux"), ...]``
        * ``label``: The label of the tag classifier. Default is ``"tag"``
        * ``default_tag``: The tag classifier to use if none was given. Default is ``"other"``
        * ``secondary_tags``: List of tuples in the format (secondary_label, secondary_default_tag). Default is empty list.

    Returns:

        Aggregated tags dictionary from ``aggregate_tagged_graph``, and tagged supply chain graph from ``recurse_tagged_database``.

    """

    lca = LCA(functional_unit, methods[0])
    lca.lci()  # factorize=True)
    lca.lcia()

    method_dicts = [{o[0]: o[1] for o in Method(method).load()} for method in methods]

    graph = [
        multi_recurse_tagged_database(
            key, amount, methods, method_dicts, lca, label, default_tag, secondary_tags
        )
        for key, amount in functional_unit.items()
    ]

    return multi_aggregate_tagged_graph(graph), graph


def multi_aggregate_tagged_graph(graph):

    """Aggregate a graph produced by ``multi_recurse_tagged_database`` by the provided tags.

    Outputs a dictionary with keys of tags and numeric values.

    Note: this only aggregates on the primary tag, secondary tags are not aggregated

    .. code-block:: python

        {'a tag': [list of summed LCIA scores with one sum per method]}

    """

    def recursor(obj, scores):
        if not scores.get(obj["tag"]):
            scores[obj["tag"]] = [x for x in obj["impact"]]
        else:
            scores[obj["tag"]] = [
                sum(x) for x in zip(scores[obj["tag"]], obj["impact"])
            ]

        for flow in obj["biosphere"]:
            if not scores.get(flow["tag"]):
                scores[flow["tag"]] = [x for x in flow["impact"]]
            else:
                scores[flow["tag"]] = [
                    sum(x) for x in zip(scores[flow["tag"]], flow["impact"])
                ]

        for exc in obj["technosphere"]:
            scores = recursor(exc, scores)
        return scores

    scores = defaultdict(int)
    for obj in graph:
        scores = recursor(obj, scores)
    return scores


def multi_recurse_tagged_database(
    activity, amount, methods, method_dicts, lca, label, default_tag, secondary_tags=[]
):

    """Traverse a foreground database and assess activities and biosphere flows by tags using multiple methods.

    Input arguments:

        * ``activity``: Activity tuple or object
        * ``amount``: float
        * ``methods``: list of LCA methods (tuples)
        * ``method_dicts``: list of dictionaries of biosphere flow tuples to CFs, e.g. ``{("biosphere", "foo"): 3}`` corresponding to methods in ``methods``
        * ``lca``: An ``LCA`` object that is already initialized, i.e. has already calculated LCI
        * ``label``: string
        * ``default_tag``: string
        * ``secondary_tags``: list of tuples in the format (secondary_label, secondary_default_tag). Default is empty list.

    Returns:

    .. code-block:: python

        {
            'activity': activity object,
            'amount': float,
            'tag': string,
            'secondary_tags': [list of strings],
            'impact': [list of floats (impact of inputs from outside foreground database) with one element per method],
            'biosphere': [{
                'amount': float,
                'impact': [list of floats with one element per method],
                'tag': string,
                'secondary_tags': [list of strings]
            }],
            'technosphere': [this data structure]
        }

    """

    if isinstance(activity, tuple):
        activity = get_activity(activity)

    inputs = list(activity.technosphere())
    inside = [exc for exc in inputs if exc.input["database"] == activity["database"]]
    outside = {
        exc.input.id: exc["amount"] * amount
        for exc in inputs
        if exc["input"][0] != activity["database"]
    }

    if outside:
        outside_scores = []
        for n, m in enumerate(methods):
            lca.switch_method(m)
            lca.redo_lcia(outside)
            outside_scores.append(lca.score)
    else:
        outside_scores = [0] * len(methods)

    return {
        "activity": activity,
        "amount": amount,
        "tag": activity.get(label) or default_tag,
        "secondary_tags": [activity.get(t[0]) or t[1] for t in secondary_tags],
        "impact": outside_scores,
        "biosphere": [
            {
                "activity": exc.input,
                "amount": exc["amount"] * amount,
                "impact": [
                    exc["amount"] * amount * method_dict.get(exc["input"], 0)
                    for method_dict in method_dicts
                ],
                "tag": exc.get(label) or activity.get(label) or default_tag,
                "secondary_tags": [
                    exc.get(t[0]) or activity.get(t[0]) or t[1] for t in secondary_tags
                ],
            }
            for exc in activity.biosphere()
        ],
        "technosphere": [
            multi_recurse_tagged_database(
                exc.input,
                exc["amount"] * amount,
                methods,
                method_dicts,
                lca,
                label,
                default_tag,
                secondary_tags,
            )
            for exc in inside
        ],
    }


def get_cum_impact(graph, max_levels=100):

    """Add cumulative impact ``cum_impact`` to each ``technosphere`` level of a tagged graph.

    This function recurses until all levels in the graph have been checked, or the ``max_levels`` cutoff is reached

    Input arguments:
        * ``graph``: A tagged supply chain graph from ``recurse_tagged_database``.
        * ``max_levels``: maximum number of graph levels to check before giving up. Default is 100.

    Returns:
         Tagged supply chain graph with additional cumulative impact ``cum_impact`` key at each ``technosphere`` level.
    """

    def cum_impact_recurse(d):

        to_return = {}
        cum_impact = 0

        for k, v in d.items():
            if k == "technosphere":
                if len(v) != 0:
                    for e in v:
                        cum_impact += e["impact"]
                        if "cum_impact" in e.keys():
                            cum_impact += e["cum_impact"]

                        if k in to_return.keys():
                            to_return[k].append(cum_impact_recurse(e))
                        else:
                            to_return[k] = [cum_impact_recurse(e)]
                else:
                    to_return[k] = []

            elif k == "biosphere":
                to_return[k] = v
                if len(v) != 0:
                    for b in v:
                        cum_impact += b["impact"]

            # elif k == 'activity':
            #    to_return[k] = str(v)

            else:
                to_return[k] = v

        to_return["cum_impact"] = cum_impact

        return to_return

    return_list = []

    for subgraph in graph:

        this_d = subgraph

        for i in range(max_levels):
            prev_d = this_d
            this_d = cum_impact_recurse(prev_d)
            if this_d == prev_d:
                break

        return_list.append(this_d)

    return return_list


def get_multi_cum_impact(graph, max_levels=100):

    """Add cumulative impact ``cum_impact`` to each ``technosphere`` level of a multi method tagged graph.

    This function recurses until all levels in the graph have been checked, or the ``max_levels`` cutoff is reached

    Input arguments:
        * ``graph``: A tagged supply chain graph from ``multi_recurse_tagged_database``.
        * ``max_levels``: maximum number of graph levels to check before giving up. Default is 100.

    Returns:
         Tagged supply chain graph with additional cumulative impact ``cum_impact`` key at each ``technosphere`` level.
    """

    def multi_cum_impact_recurse(d):

        to_return = {}
        cum_impact = [0] * len(d["impact"])

        for k, v in d.items():
            if k == "technosphere":
                if len(v) != 0:
                    for e in v:
                        cum_impact = [sum(x) for x in zip(cum_impact, e["impact"])]
                        if "cum_impact" in e.keys():
                            cum_impact = [
                                sum(x) for x in zip(cum_impact, e["cum_impact"])
                            ]

                        if k in to_return.keys():
                            to_return[k].append(multi_cum_impact_recurse(e))
                        else:
                            to_return[k] = [multi_cum_impact_recurse(e)]
                else:
                    to_return[k] = []

            elif k == "biosphere":
                to_return[k] = v
                if len(v) != 0:
                    for b in v:
                        cum_impact = [sum(x) for x in zip(cum_impact, b["impact"])]

            # elif k == 'activity':
            #    to_return[k] = str(v)

            else:
                to_return[k] = v
        to_return["cum_impact"] = cum_impact

        return to_return

    return_list = []

    for subgraph in graph:

        this_d = subgraph

        for i in range(max_levels):
            prev_d = this_d
            this_d = multi_cum_impact_recurse(prev_d)
            if this_d == prev_d:
                break

        return_list.append(this_d)

    return return_list
